import dataclasses
import enum
import getpass
import logging
import sys
from typing import Dict
from typing import Optional

LOGGER = logging.getLogger(__name__)


class LauncherPlatform(enum.Flag):
    """
    Operating Systems that a launcher may support.
    """

    linux = enum.auto()
    darwin = enum.auto()
    windows = enum.auto()

    @classmethod
    def current(cls):
        """
        Get the member corresponding to the current runtime operating system.
        """
        current = sys.platform
        if current.startswith("linux"):
            return cls.linux
        if current.startswith("darwin"):
            return cls.darwin
        if current in ("win32", "cygwin"):
            return cls.windows

        raise OSError(f"Unsupported operating system '{current}'")

    def is_linux(self) -> bool:
        return self == self.linux

    def is_mac(self) -> bool:
        return self == self.darwin

    def is_windows(self) -> bool:
        return self == self.windows


_PLATFORM_MAPPING = {
    "linux": LauncherPlatform.linux,
    "windows": LauncherPlatform.windows,
    "mac": LauncherPlatform.darwin,
}
"""
Mapping of serialized platform names to their corresponding enum instance.
"""


@dataclasses.dataclass
class LauncherContext:
    """
    Collection of system-contextual values in which a profile is being parsed.

    You can compare 2 context using the equal comparator however be aware that
    a None value (implies unset) will be considered equal to any other value.

    Example::

         ctx1 = LauncherContext(platform=LauncherPlatform.linux, user="demo")
         ctx2 = LauncherContext(platform=LauncherPlatform.linux, user=None)
         assert ctx1 == ctx2

    """

    platform: Optional[LauncherPlatform] = dataclasses.field(
        default=None,
        metadata={
            "serialized_name": "os",
            "unserialize": lambda v: _PLATFORM_MAPPING[v],
            "doc": {
                "value": f"one of ``{'``, ``'.join(_PLATFORM_MAPPING.keys())}``",
                "description": "name of the operating system",
            },
        },
    )
    """
    name of the operating system
    """

    user: Optional[str] = dataclasses.field(
        default=None,
        metadata={
            "serialized_name": "user",
            "unserialize": str,
            "doc": {
                "value": f"arbitrary string",
                "description": "name of the user",
            },
        },
    )
    """
    name of the current user
    """

    def __str__(self) -> str:
        return f"os={self.platform.name};user={self.user}"

    def __eq__(self, other):
        c = self.__class__
        if not isinstance(other, c):
            raise TypeError(f"Cannot concatenate {c} and {type(other)}")

        for field in dataclasses.fields(c):
            selfvalue = getattr(self, field.name)
            othervalue = getattr(other, field.name)
            # we discard the comparison when a field is not set
            if selfvalue is None or othervalue is None:
                continue

            if selfvalue != othervalue:
                return False

        return True

    @classmethod
    def create_from_system(cls):
        """
        Fill the instance with values retrieved from the current system context.
        """
        return cls(
            platform=LauncherPlatform.current(),
            user=getpass.getuser(),
        )


_FIELDS_MAPPING: Dict[str, dataclasses.Field] = {
    field.metadata["serialized_name"]: field
    for field in dataclasses.fields(LauncherContext)
}


_ESCAPE_CHARACTER = "%!TMP}]"


def _escape(source: str) -> str:
    return source.replace("@@", _ESCAPE_CHARACTER)


def _unescape(source: str) -> str:
    return source.replace(_ESCAPE_CHARACTER, "@")


def resolve_context_expression(source: str) -> str:
    """
    Remove the profile context expression from the given source string.
    """
    escaped = _escape(source)
    if "@" not in escaped:
        return _unescape(escaped)
    return _unescape(source.split("@")[0])


def unserialize_context_expression(source: str) -> LauncherContext:
    """
    Generate a profile context based on its serialized expression form.

    The expression take the following pattern::

        .(@key=value)*

    Where ``.`` means any character, ``()*`` means it can be repated multiple times.

    Args:
        source:
            abitrary string which contain an expression to unserialize,
            may be an empty string.

    Returns:
        a new profile context instance (that may be empty).
    """
    escaped = _escape(source)

    if "@" not in escaped:
        return LauncherContext()

    members = escaped.split("@")[1:]
    asdict = {}

    for member in members:
        member = _unescape(member)
        key, value = member.split("=")
        field = _FIELDS_MAPPING.get(key)
        if not field:
            raise KeyError(
                f"Unsupported context key name '{key}'; must one of {_FIELDS_MAPPING}"
            )
        field_name = field.name
        unserialize = field.metadata["unserialize"]
        # we explicitly allow a field to be defined multiple time,
        # its last definition will take precedence in value.
        asdict[field_name] = unserialize(value)

    return LauncherContext(**asdict)



import dataclasses
import logging

from kloch.launchers import BaseLauncherSerialized
from kloch.launchers import BaseLauncherFields
from ._dataclass import SystemLauncher

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class SystemLauncherFields(BaseLauncherFields):
    subprocess_kwargs: dict = dataclasses.field(
        default="subprocess_kwargs",
        metadata={
            "description": (
                "Mapping of kwargs to pass to the internal ``subprocess.run`` call."
            ),
            "required": False,
        },
    )


class SystemLauncherSerialized(BaseLauncherSerialized):
    source = SystemLauncher

    identifier = SystemLauncher.name

    fields = SystemLauncherFields

    summary = (
        "A simple launcher executing the given command in the default system console."
    )
    description = (
        "Useless without a command specified. The launcher will just set "
        "the given environment variables for the session, execute the command, then exit."
    )

    def validate(self):
        subprocesskw = self.fields.subprocess_kwargs
        if subprocesskw in self:
            assert isinstance(
                self[subprocesskw], dict
            ), f"'{subprocesskw}': must be a dict."
        super().validate()

    # we override for type-hint
    def unserialize(self) -> SystemLauncher:
        # noinspection PyTypeChecker
        return super().unserialize()



import logging
import shutil
import socket
import time
import uuid
from pathlib import Path
from typing import List

LOGGER = logging.getLogger(__name__)


class SessionDirectory:
    """
    Manage the filesystem directory to store data during a kloch launcher session.

    Most of the data stored is arbitrary and generated depending on the launcher.
    """

    def __init__(self, path: Path):
        self.path: Path = path
        """
        Filesystem path to an existing directory.
        """

        self.meta_session_path = self.path / "kloch.session"
        """
        A file indicating the directory is a session and which contains 
        the timestamp at which the session was created.
        """

        self.profile_path = self.path / "profile.yml"
        """
        A file which contain the merged profile used to start the session.
        """

    @property
    def identifier(self) -> str:
        return self.path.name

    @property
    def timestamp(self) -> float:
        return float(self.meta_session_path.read_text())

    @classmethod
    def initialize(cls, root: Path) -> "SessionDirectory":
        """
        Generate a new unique session directory on the filesystem.

        This function should be safe to be executed from different thread on the same machine
        at the same time.
        """
        if not root.exists():
            LOGGER.debug(f"mkdir('{root}')")
            root.mkdir()

        timestamp = time.time()
        # XXX: we generate uuid to prevent collision but we shorten it as we
        #   are already pretty safe with the timestamp.
        uuid_ = uuid.uuid4().hex[:8]
        identifier: str = f"{timestamp}-{socket.gethostname()}-{uuid_}"

        path = root / identifier
        LOGGER.debug(f"mkdir('{path}')")
        path.mkdir()

        instance = cls(path)

        LOGGER.debug(f"touch('{instance.meta_session_path}')")
        instance.meta_session_path.write_text(str(timestamp))
        return instance


def get_session_dirs(root: Path) -> List[SessionDirectory]:
    """
    Get all the session directories found in the given root directory.
    """
    return [
        SessionDirectory(path)
        for path in root.glob("*")
        if path.is_dir() and SessionDirectory(path).meta_session_path.exists()
    ]


def clean_outdated_session_dirs(root: Path, lifetime: float) -> List[Path]:
    """
    Iterate through all existing session directories and delete the one which have been created longer than the given lifetime.

    This function should handle file or dirs of the given root which are not sessions.

    Args:
        root: filesystem path to an existing directory.
        lifetime: maximum lifetime in hours of a session directory

    Returns:
        list of directories filesystem path that have been removed
    """
    # convert hours to seconds
    lifetime = lifetime * 3600
    current_time = time.time()
    minimal_lifetime = current_time - lifetime

    removed = []

    for session in get_session_dirs(root=root):

        if session.timestamp > minimal_lifetime:
            continue

        try:
            shutil.rmtree(session.path)
        except Exception as error:
            LOGGER.exception(
                f"failed to remove outdated session dir '{session.path}': {error}"
            )
            continue

        removed.append(session.path)

    return removed



import dataclasses
import logging
import subprocess
from pathlib import Path
from typing import List
from typing import Optional

from kloch.launchers import BaseLauncher


LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class SystemLauncher(BaseLauncher):
    """
    A minimal launcher that just start a subprocess with the given command.
    """

    name = ".system"

    subprocess_kwargs: dict = dataclasses.field(default_factory=dict)
    """
    Mapping of kwargs to pass to python's 'subprocess.run' call.
    """

    def execute(self, tmpdir: Path, command: Optional[List[str]] = None):
        """
        Just call subprocess.run.
        """
        _command = self.command + (command or [])

        LOGGER.debug(
            f"subprocess.run({_command}, env={self.environ}, cwd={self.cwd}, **{self.subprocess_kwargs})"
        )
        result = subprocess.run(
            _command,
            env=self.environ,
            cwd=self.cwd,
            **self.subprocess_kwargs,
        )

        return result.returncode



"""
Entities that serialize a function call made to execute a software.
"""

from ._context import LauncherContext
from ._context import LauncherPlatform

from .base import BaseLauncher
from .base import BaseLauncherSerialized
from .base import BaseLauncherFields


from .system import SystemLauncher
from .system import SystemLauncherSerialized
from .python import PythonLauncher
from .python import PythonLauncherSerialized

from ._plugins import check_launcher_plugins
from ._plugins import LoadedPluginsLaunchers
from ._plugins import load_plugin_launchers

from ._get import get_available_launchers_classes
from ._get import get_available_launchers_serialized_classes
from ._get import is_launcher_plugin

from ._serialized import LauncherSerializedDict
from ._serialized import LauncherSerializedList

_BUILTINS_LAUNCHERS = [
    BaseLauncher,
    SystemLauncher,
    PythonLauncher,
]
"""
List of launchers class implementation, including the base one.
"""

_BUILTINS_LAUNCHERS_SERIALIZED = [
    BaseLauncherSerialized,
    SystemLauncherSerialized,
    PythonLauncherSerialized,
]
"""
List of serialized launchers class implementation, including the base one.
"""



import abc
import dataclasses
import logging
import os
from pathlib import Path
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from kloch._dictmerge import refacto_dict
from kloch._dictmerge import MergeableDict
from kloch._utils import expand_envvars
from kloch._utils import patch_environ
from ._dataclass import BaseLauncher
from ... import MergeRule

LOGGER = logging.getLogger(__name__)


def _resolve_path(src_str: str) -> str:
    """
    Ensure the path is system compliant if it looks like one.
    """
    try:
        exists = Path(src_str).exists()
    except OSError:
        return src_str

    if not exists:
        return src_str

    return str(Path(src_str).resolve())


def resolve_environ(environ: Dict[str, Union[str, List[str]]]) -> Dict[str, str]:
    """
    Resolve an "environ-like" dict structure to an ``os.environ`` dict structure.
    """

    def process_pair(key: str, value: str):
        if isinstance(value, List):
            value = [_resolve_path(expand_envvars(str(path))) for path in value]
            value = os.pathsep.join(value)
        else:
            value = _resolve_path(expand_envvars(str(value)))

        # reverted by context manager, we need it so a variable defined after
        # another one can reuse that first one.
        os.environ[key] = value
        return key, value

    with patch_environ():
        return refacto_dict(environ, callback=process_pair)


class _MergeableSystemEnviron(MergeableDict):
    @classmethod
    def get_merge_rule(cls, key: str) -> MergeRule:
        # we force to always append the environ keys even if not token
        if cls.resolve_key_tokens(key) == "environ":
            return MergeRule.append
        return super().get_merge_rule(key)


def _get_mergeable_system_environ():
    return _MergeableSystemEnviron({"environ": os.environ.copy()})


# we use a dataclass over enum because we need inheritance
# noinspection PyTypeChecker
@dataclasses.dataclass(frozen=True)
class BaseLauncherFields:
    """
    A mapping of 'dataclass field name': 'corresponding key name in serialized dict'.

    Where the field name is the variable name.
    This implies you can have a different variable string value than the variable name.

    Example: ``config_params="configParams"`` is valid.

    This class is never instanced and all attributes are used at class level.
    """

    # XXX: make sure each field has a type hint when subclassed !!
    # note: fields type hint are types expected in the serialized representation.
    # note: field.metadata is used in the static documentation implying it's rst syntax.

    environ: Dict[str, Union[str, List[str]]] = dataclasses.field(
        default="environ",
        metadata={
            "description": (
                "mapping of environment variable to set before starting the environment.\n"
                "\n"
                "The value can either be a regular string or a list of string.\n"
                "The list of string has each item joined using the system path separator [2]_.\n"
                "\n"
                "- All values have environment variables expanded with ``os.expandvars`` [1]_.\n"
                "  You can escape the expansion by doubling the ``$`` like ``$$``\n"
                "- All values are turned absolute and normalized [4]_ if they are existing paths.\n"
            ),
            "required": False,
        },
    )
    merge_system_environ: bool = dataclasses.field(
        default="merge_system_environ",
        metadata={
            "description": (
                "True to implicitly merge the system environment (from the machine "
                "reading the profile) with the potentially specified ``environ`` field.\n"
                'The system environ is merged as "base" so any key specified in the ``environ`` '
                "will override it."
            ),
            "required": False,
        },
    )

    command: List[str] = dataclasses.field(
        default="command",
        metadata={
            "description": "Arbitrary list of command line arguments to call at the end of the launcher execution.",
            "required": False,
        },
    )
    cwd: str = dataclasses.field(
        default="cwd",
        metadata={
            "description": (
                'Filesystem path to an existing directory to use as "current working directory".\n'
                "\n"
                "- The path will have environment variables expanded with ``os.expandvars`` [1]_. \n"
                "  You can escape the expansion by doubling the ``$`` like ``$$``.\n"
                "  You can also use variables defined in the ``environ`` key.\n"
                "- The path is turned absolute and normalized [4]_.\n"
            ),
            "required": False,
        },
    )
    priority: int = dataclasses.field(
        default="priority",
        metadata={
            "description": (
                "How much you should privilege this launcher to be used over other launchers."
                "Higher number means higher priority."
            ),
            "required": False,
        },
    )

    @classmethod
    def iterate(cls) -> List[dataclasses.Field]:
        """
        Return all the fields defined on this dataclass
        """
        return list(dataclasses.fields(cls))


T = TypeVar("T", bound="BaseLauncherSerialized")


class BaseLauncherSerialized(MergeableDict):
    """
    A BaseLauncher instance as a serialized dict object.

    The dict structure might be invalid and might need resolving, before being
    unserialized to a valid BaseLauncher instance.
    """

    source: Type[BaseLauncher] = BaseLauncher
    """
    The source class this object serialize.
    """

    identifier: str = BaseLauncher.name
    """
    An unique name among serialized launcher that can be use in a dict key: value pair.
    """

    fields: BaseLauncherFields = BaseLauncherFields
    """
    List of all the fields that are serialized.
    """

    # Used for automatically building the static html documentation.
    # All string are in the rst syntax compatible with the static documentation builder.
    summary = (
        "An abstract launcher that whose purpose is to be merged with other launchers."
    )
    description = "This launcher is never launched and is simply merged with other launchers defined in the profile."

    def __add__(self: T, other: T) -> T:
        """
        Returns:
            new instance with deepcopied structure.
        """
        if not isinstance(other, BaseLauncherSerialized):
            raise TypeError(
                f"Cannot concatenate object of type {type(other)} with {type(self)}"
            )
        # XXX: we override so the class type is defined by the right member of the + operation
        return other.__class__(super().__add__(other))

    @abc.abstractmethod
    def validate(self):
        """
        Ensure the dict structure can be resolved to a valid structure than can be unserialized.

        Raise an exeception on any issue.
        """
        environ = self.fields.environ
        if environ in self:
            assert isinstance(self[environ], dict), f"'{environ}': must be a dict."
            for key, value in self[environ].items():
                assert isinstance(key, str), f"'{environ}': key '{key}' must be a str."

        cwd = self.fields.cwd
        if cwd in self:
            assert isinstance(self[cwd], str), f"'{cwd}': must be a str."

        command = self.fields.command
        if command in self:
            assert isinstance(self[command], list), f"'{command}': must be a list."
            for value in self[command]:
                assert isinstance(
                    value, str
                ), f"'{command}': item '{value}' must be str."

        priority = self.fields.priority
        if priority in self:
            assert isinstance(self[priority], int), f"'{priority}': must be an int."

    def resolved(self) -> Dict:
        """
        Modify the dict structure, so it can be unserialized properly.
        """
        use_system_environ = self.fields.merge_system_environ

        if self.get(use_system_environ, True):
            base = self.__class__(_get_mergeable_system_environ() + self)
            # need to disable else infinite recursion
            base[use_system_environ] = False
            resolved = base.resolved()
            return resolved

        resolved = super().resolved()

        environ = self.fields.environ
        new_environ = {}
        if environ in self:
            old_environ = self.get(environ, {})
            new_environ = {}
            new_environ.update(resolve_environ(old_environ))
            new_environ = {key: str(value) for key, value in new_environ.items()}
            resolved[environ] = new_environ

        cwd = self.fields.cwd
        if self.get(cwd):
            with patch_environ():
                os.environ.clear()
                os.environ.update(new_environ)
                resolved[cwd] = str(
                    Path((expand_envvars(self[cwd]))).absolute().resolve()
                )

        # mapping of {'serialized key': 'dataclass field name'}
        fields: Dict[str, str] = {
            getattr(self.fields, field.name): field.name
            for field in dataclasses.fields(self.fields)
        }
        # remap serialized key name to dataclass field name
        dst_dict = {fields[key]: value for key, value in resolved.items()}

        return dst_dict

    def unserialize(self) -> BaseLauncher:
        """
        Convert the dict to a valid instance.
        """
        src_dict = self.resolved()
        del src_dict[self.fields.merge_system_environ]
        return self.source.from_dict(src_dict)



"""
A simple configuration system for the Kloch runtime.
"""

import dataclasses
import logging
import os
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import yaml

from kloch.constants import Environ
from kloch._utils import expand_envvars

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


def _placeholder_caster(x: T, *args, **kwargs) -> T:
    return x


def _cast_list_split(src_str: str) -> List[str]:
    return src_str.split(",")


def _cast_path(src_str: str) -> Path:
    return Path(src_str)


def _cast_path_list_split(src_str: str) -> List[Path]:
    return [Path(path) for path in src_str.split(os.pathsep)]


# caster for yaml configs:


def _ensure_path_absolute(path: Path, absolute_root: Path) -> Path:
    if path.is_absolute():
        return path
    return Path(absolute_root, path).resolve()


def _resolve_path_env_vars(path: Path) -> Path:
    resolved = expand_envvars(str(path))
    return Path(resolved)


def _make_config_caster(caster):
    def _config_caster(v, *args, **kwargs):
        return caster(v)

    return _config_caster


def _cast_config_path(src_str: str, config_dir: Path) -> Path:
    casted = _resolve_path_env_vars(Path(src_str))
    casted = _ensure_path_absolute(casted, config_dir)
    return casted


def _cast_config_path_list(src_list: List[str], config_dir: Path) -> List[Path]:
    return [_cast_config_path(path, config_dir) for path in src_list]


@dataclasses.dataclass
class KlochConfig:
    """
    Configure kloch using a simple key/value pair dataclass system.
    """

    launcher_plugins: List[str] = dataclasses.field(
        default_factory=list,
        metadata={
            "documentation": (
                "A list of importable python module names containing new launchers to support.\n\n"
                "If specified in environment variable, this must be a comma-separated list of str like ``module1,module2,module3``"
            ),
            "config_cast": _make_config_caster(list),
            "environ": Environ.CONFIG_LAUNCHER_PLUGINS,
            "environ_cast": _cast_list_split,
        },
    )

    cli_logging_paths: List[Path] = dataclasses.field(
        default_factory=list,
        metadata={
            "documentation": (
                "Filesystem path to one or multiple log file that might exists.\n"
                "If specified all the logging will also be wrote to those files.\n"
                "The log path parent directory is created if it doesn't exists.\n\n"
                "Logs are rotated between 2 files of 262.14Kb max.\n\n"
                "If specified from the environment, it must a list of path separated "
                "by the default system path separator (windows = ``;``, linux = ``:``)"
            ),
            "config_cast": _cast_config_path_list,
            "environ": Environ.CONFIG_CLI_LOGGING_PATHS,
            "environ_cast": _cast_path_list_split,
        },
    )

    cli_logging_format: str = dataclasses.field(
        default="{levelname: <7} | {asctime} [{name}] {message}",
        metadata={
            "documentation": (
                "Formatting to use for all logged messages. See python logging module documentation.\n"
                "The tokens must use the ``{`` style."
            ),
            "config_cast": _make_config_caster(str),
            "environ": Environ.CONFIG_CLI_LOGGING_FORMAT,
            "environ_cast": str,
        },
    )

    cli_logging_default_level: Union[int, str] = dataclasses.field(
        default="INFO",
        metadata={
            "documentation": (
                "Logging level to use if None have been specified.\n"
                "Can be an int or a level name as string as long as it is understandable"
                " by ``logging.getLevelName``."
            ),
            "config_cast": _make_config_caster(str),
            "environ": Environ.CONFIG_CLI_LOGGING_DEFAULT_LEVEL,
            "environ_cast": str,
        },
    )

    cli_session_dir: Optional[Path] = dataclasses.field(
        default=None,
        metadata={
            "documentation": (
                "Filesystem path to a directory that might exists.\n"
                "The directory is used to store temporarly any file generated during the executing of a launcher.\n"
                "If not specified, a system's default temporary location is used."
            ),
            "config_cast": _cast_config_path,
            "environ": Environ.CONFIG_CLI_SESSION_PATH,
            "environ_cast": _cast_path,
        },
    )

    cli_session_dir_lifetime: float = dataclasses.field(
        default=240.0,
        metadata={
            "documentation": (
                "Amount in hours before a session directory must be deleted.\n"
                "Note the deleting is performed only the next time kloch is started so it "
                "is possible a session directory exist longer if kloch is not launched for a while."
            ),
            "config_cast": _make_config_caster(float),
            "environ": Environ.CONFIG_CLI_SESSION_LIFETIME,
            "environ_cast": float,
        },
    )

    profile_roots: List[Path] = dataclasses.field(
        default_factory=list,
        metadata={
            "documentation": (
                "Filesystem path to one or multiple directory that might exists.\n"
                "The directories contain profile valid to be discoverable.\n\n"
                "If specified from the environment, it must a list of path separated "
                "by the default system path separator (windows = ``;``, linux = ``:``)"
            ),
            "config_cast": _cast_config_path_list,
            "environ": Environ.CONFIG_PROFILE_ROOTS,
            "environ_cast": _cast_path_list_split,
        },
    )

    @classmethod
    def from_file(cls, file_path: Path) -> "KlochConfig":
        """
        Generate an instance from a serialized file.
        """
        with file_path.open("r", encoding="utf-8") as file:
            asdict: Dict = yaml.safe_load(file)

        casters = {
            field.name: field.metadata["config_cast"]
            for field in dataclasses.fields(cls)
        }
        # cast config value to expected type
        # all paths type are made absolute using the config parent dir
        asdict = {
            k: casters.get(k, _placeholder_caster)(v, file_path.parent)
            for k, v in asdict.items()
        }
        return cls(**asdict)

    @classmethod
    def from_environment(cls) -> "KlochConfig":
        """
        Generate an instance from a serialized file specified in an environment variable.
        """
        environ = os.getenv(Environ.CONFIG_PATH)

        asdict = {}
        if environ:
            base = cls.from_file(Path(environ))
            asdict = dataclasses.asdict(base)

        for field in dataclasses.fields(cls):
            env_var_name = field.metadata["environ"]
            env_var_value = os.getenv(env_var_name)
            if env_var_value is not None:
                value = field.metadata["environ_cast"](env_var_value)
                asdict[field.name] = value

        return cls(**asdict)

    @classmethod
    def get_field(cls, field_name: str) -> Optional[dataclasses.Field]:
        """
        Return the dataclass field that match the given name else None.
        """
        fields = dataclasses.fields(cls)
        field = [field for field in fields if field.name == field_name]
        return field[0] if field else None


def get_config() -> KlochConfig:
    """
    Get the current kloch configuration extracted from the environment.

    A default configuration is generated if no configuration file is specified.

    Returns:
        a new config instance
    """
    return KlochConfig.from_environment()



import contextlib
import os


@contextlib.contextmanager
def patch_environ(**environ):
    """
    Temporarily change ``os.environ`` and restore it once finished.

    Usage::

        with patch_environ():
            # the environ is left untouched by default
            os.environ.clear()
            os.environ["PATH"] = "/foo"

    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


def expand_envvars(src_str: str) -> str:
    """
    Resolve environment variable pattern in the given string.

    Using ``os.path.expandvars`` but allow escaping using ``$$``
    """
    # temporary remove escape character
    new_str = src_str.replace("$$", "##tmp##")
    # environment variable expansion
    new_str = os.path.expandvars(new_str)
    # restore escaped character
    new_str = new_str.replace("##tmp##", "$")
    return new_str



import abc
import dataclasses
from pathlib import Path
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional


@dataclasses.dataclass
class BaseLauncher:
    """
    An "abstract" dataclass that describe how to start a software environment session.
    """

    # XXX: all fields defined MUST specify a default value (else inheritance issues)
    #   instead add them to the `required_fields` class variable.

    environ: Dict[str, str] = dataclasses.field(default_factory=dict)
    """
    Mapping of environment variables to set when starting the environment.
    
    The developer is reponsible of honoring the field usage in its launcher implementation.
    """

    command: List[str] = dataclasses.field(default_factory=list)
    """
    Arbitrary list of command line arguments to call at the end of the launcher execution.
    
    The developer is reponsible of honoring the field usage in its launcher implementation.
    """

    cwd: Optional[str] = None
    """
    Current working directory.
    
    The developer is reponsible of honoring the field usage in its launcher implementation.
    """

    priority: int = 0
    """
    How much you should privilege this launcher to be used over other launchers.
    """

    required_fields: ClassVar[List[str]] = []
    """
    List of dataclass field that are required to have a non-None value when instancing.
    
    Note that your subclass must have the default field value set to None for this to work.
    
    Example::
    
          @dataclasses.dataclass
          class DemoLauncher(BaseLauncher):
              # override the BaseLauncher.environ field to make it required
              environ: Dict[str, str] = None
        
              required_fields = ["environ"]
    
    """

    name: ClassVar[str] = ".base"
    """
    A unique name among all subclasses.
    """

    def __str__(self) -> str:
        fields: List[str] = []
        for field in dataclasses.fields(self):
            name = field.name
            value = getattr(self, name)
            if name == "environ":
                asstr = "{...}" if value else "{}"
            elif name == "command":
                asstr = "[...]" if value else "[]"
            else:
                asstr = str(value)

            fields += [f"{name}={asstr}"]

        fieldsstr = ", ".join(fields)
        return f"<{self.__class__.__name__} {fieldsstr}>"

    def __post_init__(self):
        for field in dataclasses.fields(self):
            if field.name in self.required_fields and getattr(self, field.name) is None:
                raise ValueError(
                    f"Missing required field '{field.name}' for instance {self}"
                )

    @abc.abstractmethod
    def execute(self, tmpdir: Path, command: Optional[List[str]] = None) -> int:
        """
        Start the given environment and execute this python session.

        Optionally execute the given command in the environment.

        Args:
            tmpdir: filesystem path to an existing temporary directory
            command: optional list of command line arguments

        Returns:
            The exit code of the execution. 0 if successfull, else imply failure.
        """
        pass  # pragma: no cover

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the instance to a python dict object.
        """
        as_dict = dataclasses.asdict(self)
        # remove optional keys that don't have a value
        as_dict = {
            key: value
            for key, value in as_dict.items()
            if key or key in self.required_fields
        }
        return as_dict

    @classmethod
    def from_dict(cls, src_dict: Dict[str, Any]) -> "BaseLauncher":
        """
        Generate an instance from a python dict object with a specific structure.
        """
        return cls(**src_dict)



import abc
import argparse
import copy
import inspect
import logging
import logging.handlers
import re
import runpy
import sys
import tempfile
import textwrap
from pathlib import Path
from typing import List
from typing import Optional
from typing import Type

import kloch
from kloch.launchers import LauncherContext
from kloch.launchers import get_available_launchers_serialized_classes
from kloch.launchers import BaseLauncher
from kloch.launchers import BaseLauncherSerialized
from kloch.session import SessionDirectory

LOGGER = logging.getLogger(__name__)

_ARGS_USER_COMMAND_DEST = "command"


class BaseParser:
    """
    The root parser who's all subparsers use as base.

    All arguments defined here are accessible by subparsers.

    Args:
        args: user command line argument already parsed by argparse
    """

    def __init__(
        self,
        args: argparse.Namespace,
        config: kloch.KlochConfig,
        original_argv: List[str],
    ):
        self._args: argparse.Namespace = args
        self._config = config
        self._argv = original_argv

    @property
    def debug(self) -> bool:
        """
        True to execute the CLI in debug mode. Usually with more verbose logging.
        """
        return self._args.debug

    @property
    def _profile_roots(self) -> List[Path]:
        """
        One or multiple filesystem path to existing directory containing profile file.
        The paths are append to the global profile roots variable.
        """
        return [Path(path) for path in self._args.profile_roots]

    @property
    def profile_roots(self) -> List[Path]:
        """
        One or multiple filesystem path to existing directory containing profile file.
        The paths are append to the global profile roots variable.
        """
        return self._config.profile_roots + self._profile_roots

    @property
    def session_root(self) -> Optional[Path]:
        return self._config.cli_session_dir

    @abc.abstractmethod
    def execute(self):
        """
        Arbitrary code that must be executed when the user ask this command.
        """
        raise NotImplementedError()

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        """
        Configure the given argparse ArgumentParser.
        """
        parser.add_argument(
            "--debug",
            action="store_true",
            help=cls.debug.__doc__,
        )
        parser.add_argument(
            "--profile_roots",
            nargs="*",
            default=[],
            help=cls._profile_roots.__doc__,
        )
        parser.set_defaults(func=cls)

    def _get_merged_profile(
        self,
        profile_identifiers: List[str],
        context: Optional[LauncherContext],
    ):
        """
        Merge each profile with its base then merge all of them from left to right.
        """
        profile_locations = self.profile_roots

        profiles = []
        for profile_id in profile_identifiers:

            aspath = Path(profile_id)
            if aspath.exists():
                profile_paths = [aspath]
            else:
                profile_paths = kloch.get_profile_file_path(
                    profile_id,
                    profile_locations=profile_locations,
                )
            if len(profile_paths) >= 2:
                print(
                    f"ERROR | Found multiple profile with identifier '{profile_id}' "
                    f": {profile_paths}.",
                    file=sys.stderr,
                )
                sys.exit(1)

            if not profile_paths:
                print(
                    f"ERROR | No profile found with identifier '{profile_id}'.",
                    file=sys.stderr,
                )
                sys.exit(1)

            profile_path = profile_paths[0]
            LOGGER.debug(f"reading profile '{profile_path}'")
            try:
                profile = kloch.read_profile_from_file(
                    profile_path,
                    profile_locations=profile_locations,
                )
            except (
                kloch.filesyntax.ProfileAPIVersionError,
                kloch.filesyntax.ProfileInheritanceError,
            ) as error:
                print(
                    f"ERROR | '{error}'.",
                    file=sys.stderr,
                )
                sys.exit(1)

            profile = profile.get_merged_profile()
            profiles.append(profile)

        profile = profiles.pop(-1)
        for base_profile in profiles:
            profile.inherit = base_profile
            profile = profile.get_merged_profile()

        if context:
            LOGGER.debug(f"filtering profile using context {context}")
            profile.launchers = profile.launchers.get_filtered_context(context)
            profile.launchers = profile.launchers.with_context_resolved()

        return profile


class RunParser(BaseParser):
    """
    A "run" sub-command.
    """

    @property
    def launcher(self) -> str:
        """
        The name of a launcher to use in the provided environment profile.

        Only required if the profile define more than one launcher profile.
        """
        return self._args.launcher

    @property
    def profile_ids(self) -> List[str]:
        """
        One or more identifier or file paths of existing environment profile(s).

        The profiles are concatenated together from left to right.
        """
        return self._args.profile_ids

    @property
    def command(self) -> List[str]:
        """
        A command to execute in the environment that is launched by the given profile.
        """
        return self._args.command

    def _execute(self, session_dir: SessionDirectory):
        LOGGER.debug(f"session dir at '{session_dir.path}'")

        plugins_names = self._config.launcher_plugins
        plugins_names_str = (": " + ",".join(plugins_names)) if plugins_names else ""
        LOGGER.debug(f"loading {len(plugins_names)} plugin modules{plugins_names_str}")
        launcher_plugins = kloch.launchers.load_plugin_launchers(
            module_names=plugins_names,
            subclass_type=BaseLauncherSerialized,
        )
        launchers_classes = get_available_launchers_serialized_classes(launcher_plugins)

        context = LauncherContext.create_from_system()
        print(f"loading {len(self.profile_ids)} profiles ...")
        profile = self._get_merged_profile(self.profile_ids, context)

        # keep a backup of the merged profile for debugging
        LOGGER.debug(f"writing merged profile to '{session_dir.profile_path}'")
        kloch.write_profile_to_file(
            profile,
            file_path=session_dir.profile_path,
            profile_locations=self.profile_roots,
            check_valid_id=False,
            extra_comments=[
                f"auto-generated profile from argv '{' '.join(self._argv)}'",
                f"context was '{context}'",
            ],
        )

        launchers_dict = profile.launchers
        launchers_list = launchers_dict.to_serialized_list(launchers_classes)
        launchers_list = launchers_list.with_base_merged()
        if not launchers_list:
            print(
                f"ERROR | No launcher defined in profile(s) <{self.profile_ids}>",
                file=sys.stderr,
            )
            sys.exit(113)

        # the user want to use a specific launcher
        if self.launcher:
            launchers_list = [
                launcher_
                for launcher_ in launchers_list
                if launcher_.identifier == self.launcher
            ]
            if not launchers_list:
                print(
                    f"ERROR | No launcher with name <{self.launcher}> "
                    f"found in profile(s) <{self.profile_ids}>",
                    file=sys.stderr,
                )
                sys.exit(112)

        launchers: List[BaseLauncher] = []
        for seriallauncher in launchers_list:
            try:
                seriallauncher.validate()
            except AssertionError as error:
                print(
                    f"ERROR | Cannot validate launcher '{seriallauncher.identifier}' from profile '{profile.identifier}': "
                    f"{error}",
                    file=sys.stderr,
                )
                sys.exit(1)
            unserialized = seriallauncher.unserialize()
            launchers.append(unserialized)

        # try to filter launcher by priorities a first time
        if len(launchers) > 1:
            launcher_by_priority = {}
            for launcher in launchers:
                launcher_by_priority.setdefault(launcher.priority, []).append(launcher)
            priorities = sorted(list(launcher_by_priority))
            highest_priority = priorities[-1]
            launchers = launcher_by_priority[highest_priority]

        # conclude that user request / priorities were not enough to only have one launcher left
        if len(launchers) > 1:
            issues = ",".join([launcher.name for launcher in launchers])
            print(
                f"ERROR | Multiple launcher with same priority found: '{issues}'."
                f" You need to specify a launcher name with --launcher"
                f" or set launchers with a different priority.",
                file=sys.stderr,
            )
            sys.exit(111)

        launcher: BaseLauncher = launchers[0]
        command = self.command or None

        LOGGER.debug(f"executing launcher={launcher} with command={command}")
        print(f"starting launcher {launcher.name}")
        sys.exit(launcher.execute(tmpdir=session_dir.path, command=command))

    def execute(self):
        session_root = self.session_root
        if session_root is None:
            with tempfile.TemporaryDirectory(prefix=f"{kloch.__name__}") as tmp_dir:
                session_dir = SessionDirectory.initialize(Path(tmp_dir))
                return self._execute(session_dir=session_dir)
        else:
            session_dir = SessionDirectory.initialize(session_root)
            return self._execute(session_dir=session_dir)

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        super().add_to_parser(parser)
        parser.add_argument(
            "profile_ids",
            type=str,
            nargs="+",
            help=cls.profile_ids.__doc__,
        )
        parser.add_argument(
            "--launcher",
            type=str,
            help=cls.launcher.__doc__,
        )
        parser.add_argument(
            "--",
            dest=_ARGS_USER_COMMAND_DEST,
            nargs="*",
            default=[],
            help=(
                "Specify multiple argument to execute in the environment as a single command.\n"
                "MUST be the last argument as anything after is consumed."
                "If the profile already specify a command, the arguments are appended to the existing command."
            ),
        )


class ListParser(BaseParser):
    """
    A "list" sub-command.
    """

    @property
    def id_filter(self) -> Optional[str]:
        """
        A regex expression to remove profiles whose identifier doesn't match.

        Example: "knots" will match all profile starting by "knots";
        ".*beta.*" will match all profile containing "beta"
        """
        return self._args.id_filter

    def execute(self):
        profile_locations = self.profile_roots
        profile_locations_txt = [str(path) for path in profile_locations]
        print(
            f"Searching {len(profile_locations)} locations: {profile_locations_txt} ..."
        )

        profile_paths = kloch.get_all_profile_file_paths(profile_locations)
        profiles: List[kloch.EnvironmentProfile] = []

        for path in profile_paths:
            try:
                profile = kloch.read_profile_from_file(path)
            except Exception as error:
                print(f"WARNING | {path}: {error}", file=sys.stderr)
                continue
            profiles.append(profile)

        profile_ids = [profile.identifier for profile in profiles]

        if self.id_filter:
            pattern = re.compile(self.id_filter)
            original_profile_count = len(profile_ids)
            profile_ids = [
                profile_id for profile_id in profile_ids if pattern.match(profile_id)
            ]
            print(
                f"Filter <{self.id_filter}> specified, reduced listed profiles from "
                f"{original_profile_count} to {len(profile_ids)} profiles."
            )

        profile_ids_txt = ":\n- " + "\n- ".join(profile_ids) if profile_ids else "."
        message = f"Found {len(profile_ids)} valid profiles{profile_ids_txt}"
        print(message)

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        super().add_to_parser(parser)
        parser.add_argument(
            "id_filter",
            type=str,
            nargs="?",
            default=None,
            help=cls.id_filter.__doc__,
        )


class ResolveParser(BaseParser):
    """
    A "resolve" sub-command.
    """

    @property
    def profile_ids(self) -> List[str]:
        """
        One or more identifier or file paths of existing environment profile(s).

        The profiles are concatenated together from left to right.
        """
        return self._args.profile_ids

    @property
    def skip_context_filtering(self) -> bool:
        """
        Do not remove launcher that does not match the current system context.

        (disable context token resolving)
        """
        return self._args.skip_context_filtering

    def execute(self):
        context = LauncherContext.create_from_system()
        if self.skip_context_filtering:
            context = None
        profile = self._get_merged_profile(self.profile_ids, context)

        try:
            serialized = kloch.filesyntax.serialize_profile(profile)
        except kloch.filesyntax.ProfileInheritanceError as error:
            print(f"ERROR | {error}", file=sys.stderr)
            sys.exit(1)

        print(serialized)

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        super().add_to_parser(parser)
        parser.add_argument(
            "profile_ids",
            type=str,
            nargs="+",
            help=cls.profile_ids.__doc__,
        )
        parser.add_argument(
            "--skip-context-filtering",
            action="store_true",
            help=cls.skip_context_filtering.__doc__,
        )


class PythonParser(BaseParser):
    """
    A "python" sub-command.
    """

    @property
    def file_path(self) -> Path:
        """
        A filesysten path to an existing python file to execute or
        an existing directory that MUST contains a __main__.py file.
        """
        return self._args.file_path

    @property
    def user_args(self) -> List[str]:
        """
        Arbitrary nummber of command line argument passed to the python file.
        """
        return self._args.user_args

    def execute(self):
        LOGGER.debug(f"about to run '{self.file_path}' with args={self.user_args}")
        # we can set it without restoring because we sys.exit anyway
        sys.argv = [str(self.file_path)] + self.user_args

        runpy.run_path(str(self.file_path), run_name="__main__")
        sys.exit()

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        super().add_to_parser(parser)
        parser.add_argument(
            "file_path",
            type=str,
            help=cls.file_path.__doc__,
        )
        parser.add_argument(
            "user_args",
            type=str,
            nargs="*",
            help=cls.user_args.__doc__,
        )


class PluginsParser(BaseParser):
    """
    A "plugins" sub-command.
    """

    @property
    def launcher_plugins(self) -> List[str]:
        """
        Manually specify the launcher_plugins configuration key instead of using
        the default/user-generated one.

        This is a list of module names.
        """
        return self._args.launcher_plugins

    def execute(self):
        plugins_names = self.launcher_plugins or self._config.launcher_plugins
        plugins_names_str = (": " + ",".join(plugins_names)) if plugins_names else ""
        print(f"about to load {len(plugins_names)} plugin modules{plugins_names_str}")
        launcher_plugins = kloch.launchers.load_plugin_launchers(
            module_names=plugins_names,
            subclass_type=BaseLauncherSerialized,
        )

        issues = kloch.launchers.check_launcher_plugins(launcher_plugins)
        for issue in issues:
            print(f"WARNING | {issue.__class__.__name__}: {issue}", file=sys.stderr)

        launcher_classes = launcher_plugins.launchers
        if launcher_classes:
            print(f"found {len(launcher_classes)} launchers:")
            for launcher in launcher_classes:
                module_path = inspect.getfile(launcher)
                print(
                    f"- {launcher.source.__name__}: {launcher.__name__} (from '{module_path}')"
                )

        sys.exit()

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser):
        super().add_to_parser(parser)
        parser.add_argument(
            "--launcher_plugins",
            type=str,
            nargs="*",
            help=cls.launcher_plugins.__doc__,
        )


class RawFormatter(argparse.HelpFormatter):
    """
    https://stackoverflow.com/a/64102901/13806195
    """

    def _fill_text(self, text, width, indent):
        # Strip the indent from the original python definition that plagues most of us.
        text = textwrap.dedent(text)
        text = textwrap.indent(text, indent)  # Apply any requested indent.
        text = text.splitlines()  # Make a list of lines
        text = [textwrap.fill(line, width) for line in text]  # Wrap each line
        text = "\n".join(text)  # Join the lines again
        return text


def _get_parser() -> argparse.ArgumentParser:
    """
    The argparse ArgumentParser that build the CLI
    """
    parser = argparse.ArgumentParser(
        kloch.__name__,
        description=(
            "Create an environment to launch software using pre-defined configurations."
        ),
    )
    BaseParser.add_to_parser(parser)
    subparsers = parser.add_subparsers(required=True)

    subparser = subparsers.add_parser(
        "run",
        description=(
            "Launch an environment as described in the given profile.\n\n"
            'Optionally specify a command to execute after the "--" argument:\n'
            '   > %(prog)s some-profile -- echo "hello world"'
        ),
        formatter_class=RawFormatter,
    )
    RunParser.add_to_parser(subparser)

    subparser = subparsers.add_parser(
        "list",
        description="List all available profiles.",
    )
    ListParser.add_to_parser(subparser)

    subparser = subparsers.add_parser(
        "resolve",
        description=(
            "Output the given profile(s) resolved and merged to a single profile."
            "The output is machine parsable as a valid yaml file "
            "(unless you use the --debug flag)."
        ),
    )
    ResolveParser.add_to_parser(subparser)

    subparser = subparsers.add_parser(
        "python",
        description=(
            "Execute the given python file with the internal python interpreter."
        ),
    )
    PythonParser.add_to_parser(subparser)

    subparser = subparsers.add_parser(
        "plugins",
        description="List information about the currently registred plugins.",
    )
    PluginsParser.add_to_parser(subparser)
    return parser


def get_cli(
    argv: Optional[List[str]] = None,
    config: kloch.KlochConfig = None,
) -> BaseParser:
    """
    Return the command line interface generated from user arguments provided.

    Args:
        argv: command line arguments; from sys.argv if not provided
        config: the kloch config instance to use for running the cli
    """
    config = config or kloch.get_config()

    parser = _get_parser()

    argv: List[str] = copy.copy(argv) or sys.argv[1:]

    # retrieve the "--" system that allow to specify an arbitrary command to execute
    user_command = None
    if "--" in argv:
        split_index = argv.index("--")
        user_command = argv[split_index + 1 :]
        argv = argv[:split_index]

    # XXX: internal feature for the PythonLauncher. It's ok not having it documented in the CLI.
    if argv and argv[0] and Path(argv[0]).exists():
        argv.insert(0, "python")

    args = parser.parse_args(argv)
    setattr(args, _ARGS_USER_COMMAND_DEST, user_command)
    parser_class: Type[BaseParser] = args.func
    instance: BaseParser = parser_class(args, config, argv)

    # clean old sessions everytime the cli is launched
    if instance.session_root and instance.session_root.exists():
        cleaned = kloch.session.clean_outdated_session_dirs(
            root=instance.session_root,
            lifetime=config.cli_session_dir_lifetime,
        )
        LOGGER.debug(f"removed {len(cleaned)} outdated session dirs")

    return instance


def run_cli(argv: Optional[List[str]] = None):
    """
    Execute the CLI based on the provided user arguments.

    Args:
        argv: command line arguments; from sys.argv if not provided
    """
    config = kloch.get_config()
    cli = kloch.get_cli(argv, config=config)
    log_level = logging.DEBUG if cli.debug else config.cli_logging_default_level

    formatter = logging.Formatter(config.cli_logging_format, style="{")

    logging.root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(log_level)
    handler.setFormatter(formatter)
    logging.root.addHandler(handler)

    for log_path in config.cli_logging_paths:
        log_path.parent.mkdir(exist_ok=True)
        handler = logging.handlers.RotatingFileHandler(
            log_path,
            # keep in sync with config option
            maxBytes=262144,
            backupCount=1,
            encoding="utf-8",
        )
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logging.root.addHandler(handler)

    LOGGER.debug(f"starting {kloch.__name__} v{kloch.__version__}")
    LOGGER.debug(f"retrieved cli with args={cli._args}")

    sys.exit(cli.execute())



__all__ = [
    "config",
    "deepmerge_dicts",
    "EnvironmentProfile",
    "Environ",
    "filesyntax",
    "get_all_profile_file_paths",
    "get_profile_file_path",
    "get_cli",
    "get_config",
    "KlochConfig",
    "launchers",
    "MergeableDict",
    "MergeRule",
    "read_profile_from_file",
    "read_profile_from_id",
    "refacto_dict",
    "run_cli",
    "serialize_profile",
    "write_profile_to_file",
]
from .constants import Environ
from ._dictmerge import MergeableDict
from ._dictmerge import MergeRule
from ._dictmerge import refacto_dict
from ._dictmerge import deepmerge_dicts
from .config import KlochConfig
from .config import get_config
from . import config
from . import launchers
from . import filesyntax
from .filesyntax import EnvironmentProfile
from .filesyntax import serialize_profile
from .filesyntax import read_profile_from_file
from .filesyntax import read_profile_from_id
from .filesyntax import write_profile_to_file
from .filesyntax import get_profile_file_path
from .filesyntax import get_all_profile_file_paths
from .cli import get_cli
from .cli import run_cli

# keep in sync with pyproject.toml
__version__ = "0.12.0"



import copy
from typing import Dict
from typing import List
from typing import Type

from kloch import MergeableDict
from ._context import LauncherContext
from ._context import unserialize_context_expression
from ._context import resolve_context_expression
from kloch.launchers import BaseLauncherSerialized


class LauncherSerializedList(List[BaseLauncherSerialized]):
    def to_dict(self) -> Dict[str, Dict]:
        """
        Convert the list to a builtin dict structure (noc ustom class used).

        Useful for serialization or to convert to a :any:`LauncherSerializedDict` instance.
        """
        return {launcher.identifier: dict(launcher) for launcher in self}

    def with_base_merged(self) -> "LauncherSerializedList":
        """
        Get a copy of this instance with the ``.base`` launcher merged with the other launchers.

        (This implies the returned instance does NOT have a ``.base`` key anymore)

        Returns:
            new instance with deepcopied structure.
        """
        self_copy: LauncherSerializedList = copy.deepcopy(self)

        # extract the potential base that all launchers should inherit
        for launcher in self_copy:
            if launcher.__class__ is BaseLauncherSerialized:
                base_launcher = launcher
                self_copy.remove(base_launcher)
                break
        else:
            return self_copy

        return LauncherSerializedList(
            [copy.deepcopy(base_launcher) + launcher for launcher in self_copy]
        )


# revert to `MergeableDict[str, Dict]` when python-3.7 support dropped
class LauncherSerializedDict(MergeableDict):
    """
    A list of launchers instance serialized as a dict structure.

    The dict is expected to have the following root structure::

        {"manager_name1": {...}, "manager_name2": {...}, ...}

    The dict structure include tokens that need to be resolved and indicate
    how to merge 2 :obj:`LauncherSerializedDict` instances together.
    See :any:`MergeableDict` for the full documentation on tokens.
    """

    def get_filtered_context(
        self, context: LauncherContext
    ) -> "LauncherSerializedDict":
        """
        Remove all launchers that doesn't match the given context.

        Returns:
             a new deepcopied instance with possibly lesser keys.
        """
        newdict = copy.deepcopy(self)
        for launcher_identifier in self.keys():
            launcher_context = unserialize_context_expression(launcher_identifier)
            if launcher_context != context:
                del newdict[launcher_identifier]

        return newdict

    def with_context_resolved(self) -> "LauncherSerializedDict":
        """
        Merge all the same launchers with different contexts to a single launcher.
        """
        toconcatenate: List[LauncherSerializedDict] = []
        for launcher_identifier in self.keys():
            resolved = resolve_context_expression(launcher_identifier)
            newdict = LauncherSerializedDict(
                {resolved: copy.deepcopy(self[launcher_identifier])}
            )
            toconcatenate.append(newdict)

        if len(toconcatenate) == 0:
            return copy.deepcopy(self)
        elif len(toconcatenate) == 1:
            return toconcatenate[0]

        return sum(toconcatenate[1:], toconcatenate[0])

    def to_serialized_list(
        self,
        launcher_classes: List[Type[BaseLauncherSerialized]],
    ) -> LauncherSerializedList:
        """
        Convert the dict structure to a list of :any:`BaseLauncherSerialized` instances.

        Args:
            launcher_classes:
                list of launchers classes that can be possibly stored in this serialized dict.

        Raises:
            ValueError: If a launcher is serialized in this instance but is unknown.

        Returns:
            deepcopied dict structure as list of instances.
        """
        _launcher_classes = {
            launcher.identifier: launcher for launcher in launcher_classes
        }
        launchers = []
        for identifier, launcher_config in self.items():
            _identifier = self.resolve_key_tokens(identifier)
            # it's not supposed to have context expression at this stage but we safe-proof it
            _identifier = resolve_context_expression(_identifier)
            launcher_class = _launcher_classes.get(_identifier)
            if not launcher_class:
                raise ValueError(
                    f"No serialized-launcher with identifier '{_identifier}' found."
                    f"Available launchers are '{', '.join(_launcher_classes.keys())}'"
                )
            launcher = launcher_class(copy.deepcopy(launcher_config))
            launchers.append(launcher)

        return LauncherSerializedList(launchers)



import dataclasses
import logging
import subprocess
import sys
from pathlib import Path
from typing import List
from typing import Optional

from kloch.launchers import BaseLauncher


LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class PythonLauncher(BaseLauncher):
    """
    A launcher that execute the given python file with kloch's own interpreter.
    """

    python_file: str = None
    """
    Filesystem path to an existing python file.
    """

    required_fields = ["python_file"]

    name = ".python"

    def execute(self, tmpdir: Path, command: Optional[List[str]] = None):
        """
        Just call ``subprocess.run`` with ``sys.executable`` + the file path
        """
        # XXX: if packaged with nuitka, sys.executable is the built executable path,
        #   but we support this at the CLI level
        _command = [sys.executable, self.python_file]
        _command += self.command + (command or [])

        LOGGER.debug(f"subprocess.run({_command}, env={self.environ}, cwd={self.cwd})")
        result = subprocess.run(_command, env=self.environ, cwd=self.cwd)

        return result.returncode



import copy
import enum
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import TypeVar

LOGGER = logging.getLogger(__name__)


class MergeRule(enum.IntEnum):
    """
    Define how 2 python objects should be merged together.
    """

    override = enum.auto()
    append = enum.auto()
    remove = enum.auto()
    ifnotexists = enum.auto()


def refacto_dict(
    src_dict: Dict,
    callback: Callable[[Any, Any], Tuple[Any, Any]],
    recursive=True,
) -> Dict:
    """
    Iterate through all key/value pairs of the given dict and execute the given callback
    at each step which return a new key/value pair for the output dict.

    Args:
        src_dict: dict obejct ot iterate
        callback: Callable expecting 2args: key, value and should return a tuple of new_key, new_value
        recursive: True to recursively process child dict

    Returns:
        a new dict instance
    """
    new_dict = {}
    for key, value in src_dict.items():
        if isinstance(value, dict) and recursive:
            value = refacto_dict(value, callback=callback, recursive=True)
        new_key, new_value = callback(key, value)
        new_dict[new_key] = new_value
    return new_dict


def deepmerge_dicts(
    over_content: Dict[str, Any],
    base_content: Dict,
    merge_rule_callback: Optional[Callable[[str], MergeRule]] = None,
    key_resolve_callback: Optional[Callable[[str], str]] = None,
) -> Dict[str, Any]:
    """
    Recursively merge the 2 given dict "over one another".

    Intended to work best with dict having the same key/value structure.

    The merging rules are defined by the given callbacks. If no callback is provided
    the default rule is to override.

    The following object types supports the ``append`` rule:

    - `list`: with a ``.extend()`` behavior
    - `dict`: deepmerged recursively

    For any other type, the `over`'s value override the `base`'s value.
    """
    new_content = copy.deepcopy(base_content)
    merge_rule_callback = merge_rule_callback or (lambda k: MergeRule.override)
    key_resolve_callback = key_resolve_callback or (lambda k: k)

    for over_key, over_value in over_content.items():
        merge_rule = merge_rule_callback(over_key)

        # put base and over at same level by resolving both
        base_keys_resolved = {
            key_resolve_callback(bk): bk for bk, _ in base_content.items()
        }
        over_key_resolved = key_resolve_callback(over_key)

        base_key: Optional[str] = base_keys_resolved.get(over_key_resolved, None)
        base_value: Optional[Any] = base_content[base_key] if base_key else None

        if merge_rule == merge_rule.remove:
            if base_key:
                del new_content[base_key]
            continue

        if merge_rule == merge_rule.override:
            if base_key:
                del new_content[base_key]
            new_content[over_key] = over_value
            continue

        if merge_rule == merge_rule.ifnotexists:
            if base_key:
                new_content[base_key] = base_value
            else:
                new_content[over_key] = over_value
            continue

        # reaching here implies `merge_rule.append`

        if isinstance(over_value, dict) and isinstance(base_value, dict):
            new_value = deepmerge_dicts(
                over_value,
                base_content=base_value,
                merge_rule_callback=merge_rule_callback,
                key_resolve_callback=key_resolve_callback,
            )

        elif isinstance(over_value, list) and isinstance(base_value, list):
            new_value = [] + base_value + over_value

        else:
            new_value = copy.deepcopy(over_value)

        if base_key:
            # we have merged `base` value with `over` so we can remove it safely
            del new_content[base_key]
        new_content[over_key] = new_value
        continue

    return new_content


def _remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


T = TypeVar("T", bound="MergeableDict")


class MergeableDict(dict):
    """
    A dict that can be deep-merged with another dict.

    The merging algorithm is defined in :obj:`deepmerge_dicts`.

    You can define the merging granularity on a per-key basis by adding token prefix
    to your keys.

    Available tokens are found in :obj:`MergeableDict.tokens`.

    Example:

        .. exec_code::
            :caption_output: results:
            :language_output: python

            from kloch import MergeableDict
            dict_1 = MergeableDict({"+=config": {"cache": True, "level": 3, "port": "A46"}})
            dict_2 = MergeableDict({"+=config": {"cache": False, "-=level": 3}})
            print(dict_1 + dict_2)

    """

    class tokens:
        append = "+="
        remove = "-="
        override = "=="
        ifnotexists = "!="

    def __add__(self: T, other: T) -> T:
        """
        Merge the 2 dict structure together with ``other`` merged over ``self``.

        The returned class type is of the left member of the + operation.

        Returns:
            new instance with deepcopied structure.
        """
        if not isinstance(other, MergeableDict):
            raise TypeError(
                f"Cannot concatenate object of type {type(other)} with {type(self)}"
            )

        new_content = deepmerge_dicts(
            over_content=other,
            base_content=self,
            key_resolve_callback=self.resolve_key_tokens,
            merge_rule_callback=self.get_merge_rule,
        )
        return self.__class__(new_content)

    @classmethod
    def resolve_key_tokens(cls, key: str) -> str:
        """
        Ensure the given key has all potential tokens removed.
        """
        resolved = key
        for token in [
            cls.tokens.append,
            cls.tokens.remove,
            cls.tokens.override,
            cls.tokens.ifnotexists,
        ]:
            resolved = _remove_prefix(resolved, token)
        return resolved

    @classmethod
    def get_merge_rule(cls, key: str) -> MergeRule:
        """
        Extract the :obj:`MergeRule` for the given key based on its potential token.
        """
        if key.startswith(cls.tokens.append):
            return MergeRule.append
        if key.startswith(cls.tokens.remove):
            return MergeRule.remove
        if key.startswith(cls.tokens.override):
            return MergeRule.override
        if key.startswith(cls.tokens.ifnotexists):
            return MergeRule.ifnotexists
        return MergeRule.append

    def get(self, key, default=None, ignore_tokens: bool = False):
        """
        Args:
            key: key's value to retrieve
            default: value to return if key is not in the dict
            ignore_tokens:
                if True, both key and self are used with a resolved variant (without tokens).
                For example ``.get("config",ignore_tokens=True)`` would still return
                ``{...}`` if ``self=={"+=config":{...}}``.
        """

        new_key = key
        if ignore_tokens:
            new_key = self.resolve_key_tokens(key)
            resolved_mapping = {
                self.resolve_key_tokens(child_key): child_key for child_key in self
            }
            if new_key in resolved_mapping:
                new_key = resolved_mapping[new_key]
            else:
                new_key = key

        return super().get(new_key, default)

    def resolved(self) -> Dict:
        """
        Get the dict structure with all tokens resolved.

        Without tokens, the returned object become a regular dict instance.

        Returns:
            deepcopied dict structure.
        """

        def process_pair(key: str, value: str):
            new_key = self.resolve_key_tokens(key)
            return new_key, value

        new_content = refacto_dict(
            src_dict=copy.deepcopy(self),
            callback=process_pair,
            recursive=True,
        )
        return dict(new_content)



"""
We define a simple config system that describe how to build a software environment using
different "software launchers".

The config system can handle the merging of 2 configs structure.
"""

import copy
import dataclasses
from typing import Any
from typing import Dict
from typing import Optional

from kloch.launchers import LauncherSerializedDict


@dataclasses.dataclass
class EnvironmentProfile:
    """
    A profile is a collection of parameters required to start a pre-defined launcher.

    This can be seen as the context/environment necessary to run a launcher thus its
    full name 'Environment Profile' that we abbreviate to profile for convenience.

    Profiles can inherit each other by specifying a `inherit` attribute. The inheritance
    only merge the content of the ``launchers`` attribute between 2 profiles.
    """

    identifier: str
    version: str
    inherit: Optional["EnvironmentProfile"]
    launchers: LauncherSerializedDict

    @classmethod
    def from_dict(cls, serialized: Dict) -> "EnvironmentProfile":
        """
        Generate a profile instance from a serialized dict object.

        No type checking is performed and the user is reponsible for the correct
        type being stored in the dict.
        """
        identifier: str = serialized["identifier"]
        version: str = serialized["version"]
        inherit: Optional["EnvironmentProfile"] = serialized.get("inherit", None)
        launchers: LauncherSerializedDict = serialized["launchers"]

        return EnvironmentProfile(
            identifier=identifier,
            version=version,
            inherit=inherit,
            launchers=launchers,
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert a profile instance to a serialized dict object.
        """
        serialized = {
            "identifier": self.identifier,
            "version": self.version,
        }
        if self.inherit:
            serialized["inherit"] = self.inherit

        serialized["launchers"] = copy.deepcopy(self.launchers)
        return serialized

    def get_merged_profile(self):
        """
        Resolve the inheritance the profile might have over another profile.

        Returns:
            a new instance.
        """
        launchers = self.launchers
        if self.inherit:
            launchers = self.inherit.get_merged_profile().launchers + launchers

        return EnvironmentProfile(
            identifier=self.identifier,
            version=self.version,
            inherit=None,
            launchers=launchers,
        )



import dataclasses
import importlib
import inspect
import logging
from typing import Dict
from typing import Generic
from typing import List
from typing import Type
from typing import TypeVar

from kloch.launchers import BaseLauncher
from kloch.launchers import BaseLauncherSerialized

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


@dataclasses.dataclass
class LoadedPluginsLaunchers(Generic[T]):
    launchers: List[T]
    """
    List of external plugins that have been loaded
    """

    missed: Dict[str, str]
    """
    Given plugins that couldn't be loaded as mapping of {"module name": "error message"}
    """

    given: List[str]
    """
    Original list of modules name the plugins were extracted from.
    """


def load_plugin_launchers(
    module_names: List[str],
    subclass_type: Type[T],
) -> LoadedPluginsLaunchers[Type[T]]:
    """
    Retrieve the launcher subclasses from the given module names.

    Import error are silenced and stored in the returned object.

    Args:
        module_names: list of importable python module names
        subclass_type: base class of the launchers to return

    Returns:
        An instance of plugins loaded
    """
    plugins = []
    missed = {}

    for module_name in module_names:
        try:
            module = importlib.import_module(module_name)
        except (ModuleNotFoundError, ImportError) as error:
            missed[module_name] = str(error)
            continue

        module_content = inspect.getmembers(module, inspect.isclass)
        module_launchers = [
            obj[1]
            for obj in module_content
            if issubclass(obj[1], subclass_type) and not obj[1] is subclass_type
        ]
        if not module_launchers:
            missed[module_name] = (
                f"Module doesn't have any subclass of '{subclass_type}'"
            )
            continue

        for launcher in module_launchers:
            if launcher in plugins:
                LOGGER.warning(
                    f"Got duplicated plugin '{launcher}' from module '{module_name}': skipping"
                )
                continue
            plugins.append(launcher)

    return LoadedPluginsLaunchers(
        launchers=plugins,
        missed=missed,
        given=module_names.copy(),
    )


def _assert(condition: bool, message: str):
    if not condition:
        raise AssertionError(message)


def _check_launcher_implementation(launcher: Type[BaseLauncher]):
    """
    Raise an AssertionError if the given launcher subclass is not properly implemented.

    Args:
        launcher: a BaseLauncher class/subclass to check
    """
    if launcher is not BaseLauncher:
        _assert(
            launcher.name != BaseLauncher.name,
            f"Implementation '{launcher.__name__}' forgot to override 'name' attribute.",
        )

    _assert(
        issubclass(launcher, BaseLauncher),
        f"Implementation '{launcher.__name__}' must be a direct subclass of '{BaseLauncher.__name__}'.",
    )


def _check_launcher_serialized_implementation(launcher: Type[BaseLauncherSerialized]):
    """
    Raise an AssertionError if the given serialized launcher subclass is not properly implemented.

    Args:
        launcher: a BaseLauncherSerialized class/subclass to check
    """
    _assert(
        issubclass(launcher, BaseLauncherSerialized),
        f"Implementation '{launcher.__name__}' must be a direct subclass of '{BaseLauncherSerialized.__name__}'.",
    )

    def missing(field: str) -> str:
        return (
            f"Implementation '{launcher.__name__}' is missing the '{field}' attribute."
        )

    _assert(hasattr(launcher, "fields"), missing("fields"))
    _assert(hasattr(launcher, "source"), missing("source"))
    _assert(hasattr(launcher, "identifier"), missing("identifier"))
    _assert(hasattr(launcher, "summary"), missing("summary"))
    _assert(hasattr(launcher, "description"), missing("description"))

    def wrong_override(field: str) -> str:
        return f"Implementation '{launcher.__name__}' doesn't override the base '{field}' attribute."

    if launcher is not BaseLauncherSerialized:
        _assert(
            launcher.source is not BaseLauncherSerialized,
            wrong_override("source"),
        )
        _assert(
            launcher.fields is not BaseLauncherSerialized.fields,
            wrong_override("fields"),
        )
        _assert(
            launcher.summary is not BaseLauncherSerialized.summary,
            wrong_override("summary"),
        )
        _assert(
            launcher.description is not BaseLauncherSerialized.description,
            wrong_override("description"),
        )
        _assert(
            launcher.identifier is not BaseLauncherSerialized.identifier,
            wrong_override("identifier"),
        )

    fields_ser = dataclasses.fields(launcher.fields)
    fields_source = dataclasses.fields(launcher.source)

    _assert(
        len(fields_ser) >= len(fields_source),
        f"Implementation '{launcher.__name__}' must define as much or more fields than its source '{launcher.source}'",
    )

    for field_ser in fields_ser:
        _assert(
            "required" in field_ser.metadata,
            f"Implementation '{launcher.__name__}' field '{field_ser.name}' is missing a 'required' attribute in its metadata.",
        )
        _assert(
            "description" in field_ser.metadata,
            f"Implementation '{launcher.__name__}' field '{field_ser.name}' is missing a 'description' attribute in its metadata.",
        )


class PluginModuleError(Exception):
    pass


class PluginImplementationError(Exception):
    pass


def check_launcher_plugins(
    plugins: LoadedPluginsLaunchers[Type[BaseLauncherSerialized]],
) -> List[Exception]:
    """
    Return any issue/error the given launcher may have.

    This function is not supposed to raise if a launcher is invalid.

    Args:
        plugins: collection of plugin launcher loaded from external modules.

    Returns:
        list of errors found in given loaded plugins, empty if no errors.
    """
    errors: List[Exception] = []
    if plugins.missed:
        for module, error in plugins.missed.items():
            error = PluginModuleError(f"Module '{module}' was not loaded: {error}")
            errors.append(error)

    for launcher in plugins.launchers:
        try:
            _check_launcher_serialized_implementation(launcher)
        except AssertionError as error:
            error = PluginImplementationError(str(error))
            errors.append(error)

    launchers = [launcher.source for launcher in plugins.launchers]

    try:
        for launcher in launchers:
            _check_launcher_implementation(launcher)
    except AssertionError as error:
        error = PluginImplementationError(str(error))
        errors.append(error)

    return errors



"""
Serialization and unserialization of environment profile to disk.
"""

__all__ = [
    "EnvironmentProfile",
    "ProfileInheritanceError",
    "ProfileAPIVersionError",
    "ProfileIdentifierError",
    "is_file_environment_profile",
    "get_profile_file_path",
    "get_all_profile_file_paths",
    "read_profile_from_file",
    "read_profile_from_id",
    "serialize_profile",
    "write_profile_to_file",
]

from ._profile import EnvironmentProfile
from ._io import ProfileInheritanceError
from ._io import ProfileAPIVersionError
from ._io import ProfileIdentifierError
from ._io import is_file_environment_profile
from ._io import get_profile_file_path
from ._io import get_all_profile_file_paths
from ._io import read_profile_from_file
from ._io import read_profile_from_id
from ._io import serialize_profile
from ._io import write_profile_to_file



from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import kloch.launchers
from kloch.launchers import BaseLauncher
from kloch.launchers import BaseLauncherSerialized
from ._plugins import LoadedPluginsLaunchers


T = TypeVar("T")


def _collect_launchers(
    natives: List[Type[T]],
    plugins: Optional[LoadedPluginsLaunchers[Type[T]]] = None,
) -> List[Type[T]]:
    """
    Combine given launchers to flat list of unique objects.
    """
    plugins = plugins.launchers if plugins else []
    launchers = natives + plugins
    # remove duplicates but preserve order
    return [
        launcher
        for index, launcher in enumerate(launchers)
        if launcher not in launchers[:index]
    ]


def get_available_launchers_classes(
    plugins: Optional[LoadedPluginsLaunchers[Type[BaseLauncher]]] = None,
) -> List[Type[BaseLauncher]]:
    """
    Collect all the launchers classes available.

    This is simply the given plugin launchers + builtin launchers.

    Args:
        plugins: collection of launcher loaded from plugins.
    """
    # noinspection PyProtectedMember
    return _collect_launchers(
        natives=kloch.launchers._BUILTINS_LAUNCHERS.copy(),
        plugins=plugins,
    )


def get_available_launchers_serialized_classes(
    plugins: Optional[LoadedPluginsLaunchers[Type[BaseLauncherSerialized]]] = None,
) -> List[Type[BaseLauncherSerialized]]:
    """
    Collect all the serialized launchers classes available.

    This is simply the given plugin launchers + builtin launchers.

    Args:
        plugins: collection of launcher loaded from plugins
    """
    # noinspection PyProtectedMember
    return _collect_launchers(
        natives=kloch.launchers._BUILTINS_LAUNCHERS_SERIALIZED.copy(),
        plugins=plugins,
    )


# noinspection PyProtectedMember
def is_launcher_plugin(
    launcher: Union[Type[BaseLauncher], Type[BaseLauncherSerialized]]
) -> bool:
    """
    Return True if the given launcher is an external plugin else False if builtin.
    """
    if launcher in kloch.launchers._BUILTINS_LAUNCHERS:
        return False
    return launcher not in kloch.launchers._BUILTINS_LAUNCHERS_SERIALIZED



"""
Variables whose values doesn't change during application runtime.
"""

from typing import List

_KLOCH_CONFIG_PREFIX = "KLOCH_CONFIG"


class Environ:
    """
    Environment variables used across the application.
    """

    CONFIG_PATH = f"{_KLOCH_CONFIG_PREFIX}_PATH"
    """
    Environment variable that must specify a file path to an existing configuration file.
    """

    CONFIG_LAUNCHER_PLUGINS = f"{_KLOCH_CONFIG_PREFIX}_launcher_plugins".upper()

    CONFIG_CLI_LOGGING_PATHS = f"{_KLOCH_CONFIG_PREFIX}_cli_logging_paths".upper()

    CONFIG_CLI_LOGGING_FORMAT = f"{_KLOCH_CONFIG_PREFIX}_cli_logging_format".upper()

    CONFIG_CLI_LOGGING_DEFAULT_LEVEL = (
        f"{_KLOCH_CONFIG_PREFIX}_cli_logging_default_level".upper()
    )

    CONFIG_CLI_SESSION_PATH = f"{_KLOCH_CONFIG_PREFIX}_cli_session_path".upper()

    CONFIG_CLI_SESSION_LIFETIME = f"{_KLOCH_CONFIG_PREFIX}_cli_lifetime".upper()

    CONFIG_PROFILE_ROOTS = f"{_KLOCH_CONFIG_PREFIX}_profile_roots".upper()

    @classmethod
    def list_all(cls) -> List[str]:
        """
        Return all environment variables supported by the application.
        """
        return [
            obj
            for obj_name, obj in vars(cls).items()
            if not obj_name.startswith("__")
            and isinstance(obj, str)
            and obj.startswith("KLOCH")
        ]



import dataclasses
import logging
import os
from pathlib import Path
from typing import Dict
from typing import List

from kloch.launchers import BaseLauncherSerialized
from kloch.launchers import BaseLauncherFields
from ._dataclass import PythonLauncher

LOGGER = logging.getLogger(__name__)


# noinspection PyTypeChecker
@dataclasses.dataclass(frozen=True)
class PythonLauncherFields(BaseLauncherFields):
    python_file: str = dataclasses.field(
        default="python_file",
        metadata={
            "description": (
                "Filesystem path to an existing python file.\n"
                "The path will have environment variables expanded with ``os.expandvars`` [1]_.\n"
                "The path is turned absolute and normalized. [4]_\n"
            ),
            "required": True,
        },
    )
    # we override just for the metadata attribute
    # noinspection PyDataclass
    command: List[str] = dataclasses.field(
        default=BaseLauncherFields.command,
        metadata={
            "description": "Arbitrary list of command line arguments passed to the python file.",
            "required": False,
        },
    )


class PythonLauncherSerialized(BaseLauncherSerialized):
    source = PythonLauncher

    identifier = PythonLauncher.name

    fields = PythonLauncherFields

    summary = (
        "A launcher that execute the given python file with kloch's own interpreter."
    )
    description = (
        "Execute the given python file with kloch internal python interpreter.\n"
        "\n"
        "All ``command`` keys are passed as args to the python script."
    )

    def validate(self):
        super().validate()
        python_file = self.fields.python_file
        assert python_file in self, f"'{python_file}': missing or empty attribute."
        assert isinstance(self[python_file], str), f"'{python_file}': must be a str."

    def resolved(self) -> Dict:
        resolved = super().resolved()
        python_file = self.fields.python_file

        old_python_file = self[python_file]
        new_python_file = Path(os.path.expandvars(old_python_file)).absolute().resolve()
        resolved[python_file] = str(new_python_file)

        return resolved

    # we override for type-hint
    def unserialize(self) -> PythonLauncher:
        # noinspection PyTypeChecker
        return super().unserialize()



import logging
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional

import yaml

from ._profile import LauncherSerializedDict
from ._profile import EnvironmentProfile


LOGGER = logging.getLogger(__name__)


KENV_PROFILE_MAGIC = "kloch_profile"
KENV_PROFILE_VERSION = 4


class ProfileAPIVersionError(Exception):
    """
    Issue with the '__magic__' attribute of a profile.
    """

    pass


class ProfileInheritanceError(Exception):
    """
    Issue with the 'inherit' attribute of a profile.
    """

    pass


class ProfileIdentifierError(Exception):
    """
    Issue with the 'identifier' attribute of a profile.
    """

    pass


def is_file_environment_profile(file_path: Path) -> bool:
    """
    Return True if the given file is an Environment Profile.

    Args:
        file_path: filesystem path to an existing file
    """
    if not file_path.suffix == ".yml":
        return False

    with file_path.open("r", encoding="utf-8") as file:
        content = yaml.safe_load(file)

    if not content:
        return False

    return content.get("__magic__", "").startswith(KENV_PROFILE_MAGIC)


def get_all_profile_file_paths(locations: Optional[List[Path]] = None) -> List[Path]:
    """
    Get all the environment-profile file paths as registred by the user.

    Args:
        locations: list of filesystem path to directory that might exist
    """
    locations = locations or []
    return [
        path
        for location in locations
        for path in location.glob("*.yml")
        if is_file_environment_profile(path)
    ]


def _get_profile_identifier(file_path: Path) -> str:
    with file_path.open("r", encoding="utf-8") as file:
        asdict: Dict = yaml.safe_load(file)
    return asdict["identifier"]


def get_profile_file_path(
    profile_id: str,
    profile_locations: Optional[List[Path]] = None,
) -> List[Path]:
    """
    Get the filesystem location to the profile(s) with the given name.

    Args:
        profile_id: identifier that must match returned profiles.
        profile_locations:
            list of filesystem path to potential existing directories containing profiles.

    Returns:
        list of filesystem path to existing files . Might be empty.
    """
    profile_paths = get_all_profile_file_paths(locations=profile_locations)
    profiles: List[Path] = [
        path for path in profile_paths if _get_profile_identifier(path) == profile_id
    ]
    return profiles


def read_profile_from_file(
    file_path: Path,
    profile_locations: Optional[List[Path]] = None,
) -> EnvironmentProfile:
    """
    Generate an instance from a serialized file on disk.

    Raises:
        ProfileAPIVersionError:
        ProfileInheritanceError:

    Args:
        file_path:
            filesystem path to an existing valid profile file.
        profile_locations:
            list of filesystem path to potential existing directories containing profiles.
    """
    with file_path.open("r", encoding="utf-8") as file:
        asdict: Dict = yaml.safe_load(file)

    profile_version = int(asdict["__magic__"].split(":")[-1])
    if not profile_version == KENV_PROFILE_VERSION:
        raise ProfileAPIVersionError(
            f"Cannot read profile with version <{profile_version}> while current "
            f"API version is <{KENV_PROFILE_VERSION}>."
        )
    del asdict["__magic__"]

    super_name: Optional[str] = asdict.get("inherit", None)
    if super_name:
        super_paths = get_profile_file_path(
            super_name,
            profile_locations=profile_locations,
        )
        if len(super_paths) >= 2:
            raise ProfileInheritanceError(
                f"Found multiple profile with identifier '{super_name}' "
                f"specified from profile '{file_path}': {super_paths}."
            )
        if not super_paths:
            raise ProfileInheritanceError(
                f"No profile found with identifier '{super_name}' "
                f"specified from profile '{file_path}'."
            )

        super_profile = read_profile_from_file(file_path=super_paths[0])
        asdict["inherit"] = super_profile

    launchers = LauncherSerializedDict(asdict["launchers"])
    asdict["launchers"] = launchers

    profile = EnvironmentProfile.from_dict(asdict)
    return profile


def read_profile_from_id(
    profile_id: str,
    profile_locations: Optional[List[Path]] = None,
) -> EnvironmentProfile:
    """
    Generate a profile instance from a serialized file on disk retrieved using the given identifier.

    Raises error if the profile file is not built properly.

    This a convenient function wrapping :func:`read_profile_from_file` and
    :func:`get_profile_file_path` and assuming that no profile with the same
    identifier exist in the locations.

    Args:
        profile_id: identifier that must match the profile.
        profile_locations:
            list of filesystem path to potential existing directories containing profiles.

    Returns:
        a profile instance
    """
    profile_paths = get_profile_file_path(
        profile_id=profile_id,
        profile_locations=profile_locations,
    )
    profile = read_profile_from_file(
        file_path=profile_paths[0],
        profile_locations=profile_locations,
    )
    return profile


def serialize_profile(
    profile: EnvironmentProfile,
    profile_locations: Optional[List[Path]] = None,
) -> str:
    """
    Convert the instance to a serialized dictionnary intended to be written on disk.

    Raises:
        ProfileInheritanceError: if the inherited profile specified is not found on disk
    """
    asdict = {"__magic__": f"{KENV_PROFILE_MAGIC}:{KENV_PROFILE_VERSION}"}
    asdict.update(profile.to_dict())

    super_profile: Optional[EnvironmentProfile] = asdict.get("inherit", None)
    if super_profile:
        super_path = get_profile_file_path(
            profile_id=super_profile.identifier,
            profile_locations=profile_locations,
        )
        if not super_path:
            raise ProfileInheritanceError(
                f"Profile '{super_profile.identifier}' specified for inheritance on "
                f"profile '{profile.identifier}' cannot be found on disk."
            )
        asdict["inherit"] = super_profile.identifier

    # remove custom class wrapper
    asdict["launchers"] = dict(asdict["launchers"])

    return yaml.dump(asdict, sort_keys=False)


def write_profile_to_file(
    profile: EnvironmentProfile,
    file_path: Path,
    profile_locations: Optional[List[Path]] = None,
    check_valid_id: bool = True,
    extra_comments: List[str] = None,
) -> Path:
    """
    Convert the instance to a serialized file on disk.

    Raises:
        ProfileIdentifierError: if check_valid_id=True and the profile identifier is not unique

    Args:
        profile: profile instance to write to disk
        file_path:
            filesystem path to a file that might exist.
            parent location is expected to exist.
        check_valid_id:
            if True, ensure the identifier of the profile is unique among all ``profile_locations``
        profile_locations:
            list of filesystem path to potential existing directories containing profiles.
        extra_comments: optional lines of comments to put in the yaml header
    """
    if check_valid_id:
        profile_paths = get_profile_file_path(
            profile_id=profile.identifier,
            profile_locations=profile_locations,
        )
        if profile_paths and file_path not in profile_paths:
            raise ProfileIdentifierError(
                f"Found multiple profile with identifier '{profile.identifier}'."
            )

    serialized = serialize_profile(profile, profile_locations=profile_locations)

    extra_comments = extra_comments or []
    extra_comments = "# " + "\n# ".join(extra_comments)
    serialized = extra_comments + "\n" + serialized

    file_path.write_text(serialized)
    return file_path
